/*
 * Copyright (C) 2005 Jordan Kiang
 * jordan-at-kiang.org
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package tetris;

/**
 * A thread to drop a piece at a given rate.
 * It doesn't seem to be possible to restart a Thread in CLDC 1.0.
 * Instead we just instantiate a new DropThread when needed (ie after a pause).
 * 
 * SkipNextTick stuff is a probably not ideal.  Ideally after a quick drop the tick would be reset
 *   so that exactly one tick cycle occurs before the piece drops a row, rather than a remainder of a cycle
 *   or 1 cycle + a remainder of a cycle.  Don't really see a clean way of doing this, however, using
 *   limited thread functionality in CLDC 1.0.
 */
public class DropThread extends Thread {
	
	private TetrisMidlet game;	// the game using this thread
	
	private boolean running;		// is the thread running
	private boolean skipNextTick;	// whether the thread should skip dropping the piece on the next tick
									// this is used after a quick drop so that the new piece doesn't just
									// drop after the remaining period of time in the current tick
	
	/**
	 * Make a new drop thread.
	 * @param game the game midlet
	 */
	public DropThread(TetrisMidlet game) {
		this.game = game;
		this.running = true;
		this.skipNextTick = true;	// skip the first drop, because ticking is at the start of the loop,
									// and we don't want an immediate drop when the thread is started
	}
	
	/**
	 * Stop the thread.  This will cause this thread to exit its running loop.
	 * Apparently it can't be restarted, and will instead a new instance needs to be used in the future.
	 */
	public void stopThread() {
		this.running = false;
	}
	
	/**
	 * Skip the next tick, no dropping action will be taken at the next tick.
	 */
	public void skipNextTick() {
		this.skipNextTick = true;
	}
	
	/**
	 * The drop running loop.
	 */
	public void run() {
	    while(this.running) {			
			// drop the active piece
			if(this.skipNextTick) {
				this.skipNextTick = false;
			} else {
				this.game.tick();
			}
			
			try {
				Thread.sleep(this.game.getTickSpeed());	// sleep between tick
			} catch(InterruptedException ie) {
			    // who's interrupting us? just drop out of running
			}
		}
	}
}
